/*
 *      T T C P . C
 *
 * Test TCP connection.  Makes a connection on port 5001
 * and transfers fabricated buffers or data copied from stdin.
 *
 * Usable on 4.2, 4.3, and 4.1a systems by defining one of
 * BSD42 BSD43 (BSD41a)
 * Machines using System V with BSD sockets should define SYSV.
 *
 * Modified for operation under 4.2BSD, 18 Dec 84
 *      T.C. Slattery, USNA
 * Minor improvements, Mike Muuss and Terry Slattery, 16-Oct-85.
 * Modified in 1989 at Silicon Graphics, Inc.
 *      catch SIGPIPE to be able to print stats when receiver has died 
 *      for tcp, don't look for sentinel during reads to allow small transfers
 *      increased default buffer size to 8K, nbuf to 2K to transfer 16MB
 *      moved default port to 5001, beyond IPPORT_USERRESERVED
 *      make sinkmode default because it is more popular, 
 *              -s now means don't sink/source 
 *      count number of read/write system calls to see effects of 
 *              blocking from full socket buffers
 *      for tcp, -D option turns off buffered writes (sets TCP_NODELAY sockopt)
 *      buffer alignment options, -A and -O
 *      print stats in a format that's a bit easier to use with grep & awk
 *      for SYSV, mimic BSD routines to use most of the existing timing code
 * Modified by Steve Miller of the University of Maryland, College Park
 *      -b sets the socket buffer size (SO_SNDBUF/SO_RCVBUF)
 * Modified Sept. 1989 at Silicon Graphics, Inc.
 *      restored -s sense at request of tcs@brl
 * Modified Oct. 1991 at Silicon Graphics, Inc.
 *      use getopt(3) for option processing, add -f and -T options.
 *      SGI IRIX 3.3 and 4.0 releases don't need #define SYSV.
 *
 * Distribution Status -
 *      Public Domain.  Distribution Unlimited.
 */
#ifndef lint
static char RCSid[] = "ttcp.c $Revision: 1.1 $";
#endif

#define BSD43
/* #define BSD42 */
/* #define BSD41a */
/* #define SYSV */      /* required on SGI IRIX releases before 3.3 */

#include <stdio.h>
#include <io.h>
//#include <signal.h>
#include <ctype.h>
#include <errno.h>
/* #include <machine/endian.h> */
#include <sys/types.h>
#include <winsock.h>
#include <time.h>           /* struct timeval */
#include <winsock.h>
#include <time.h>           /* struct timeval */

//#include <sys/times.h>
//#include <sys/param.h>
struct rusage {
    struct timeval ru_utime, ru_stime;
};
#define RUSAGE_SELF 0
//#include <sys/resource.h>

struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;

int domain, fromlen;
SOCKET fd;                         /* fd of network socket */

int buflen = 8 * 1024;          /* length of buffer */
char *buf;                      /* ptr to dynamic buffer */
int nbuf = 2 * 1024;            /* number of buffers to send in sinkmode */

int bufoffset = 0;              /* align buffer to this */
int bufalign = 16*1024;         /* modulo this */

int udp = 0;                    /* 0 = tcp, !0 = udp */
int options = 0;                /* socket options */
int one = 1;                    /* for 4.3 BSD style setsockopt() */
short port = 5001;              /* TCP port number */
char *host;                     /* ptr to name of host */
int trans;                      /* 0=receive, !0=transmit mode */
int sinkmode = 0;               /* 0=normal I/O, !0=sink/source mode */
int verbose = 0;                /* 0=print basic info, 1=print cpu rate, proc
                                 * resource usage. */
int atm = 0;
int nodelay = 0;                /* set TCP_NODELAY socket option */
int b_flag = 0;                 /* use mread() */
int sockbufsize = 0;            /* socket buffer size to use */
char fmt = 'K';                 /* output format: k = kilobits, K = kilobytes,
                                 *  m = megabits, M = megabytes, 
                                 *  g = gigabits, G = gigabytes */
int touchdata = 0;              /* access data after reading */

struct hostent *addr;
extern int errno;
int optind = 1;
char *optarg;

char Usage[] = "\
Usage: ttcp -t [-options] host [ < in ]\n\
       ttcp -r [-options > out]\n\
Common options:\n\
        -l ##   length of bufs read from or written to network (default 8192)\n\
        -u      use UDP instead of TCP\n\
        -p ##   port number to send to or listen at (default 5001)\n\
        -s      -t: source a pattern to network\n\
                -r: sink (discard) all data from network\n\
        -A      align the start of buffers to this modulus (default 16384)\n\
        -O      start buffers at this offset from the modulus (default 0)\n\
        -v      verbose: print more statistics\n\
        -M      show ATM statistics\n\
        -d      set SO_DEBUG socket option\n\
        -b ##   set socket buffer size (if supported)\n\
        -f X    format for rate: k,K = kilo{bit,byte}; m,M = mega; g,G = giga\n\
                (in actual units)  x (10**3 bit)       y (10**6)   z (10**9) \n\
Options specific to -t:\n\
        -n##    number of source bufs written to network (default 2048)\n\
        -D      don't buffer TCP writes (sets TCP_NODELAY socket option)\n\
Options specific to -r:\n\
        -B      for -s, only output full blocks as specified by -l (for TAR)\n\
        -T      \"touch\": access each byte as it's read\n\
";      

char stats[128];
unsigned long nbytes;           /* bytes on net */
unsigned long numCalls;         /* # of I/O system calls */
double elapsed_time;             /* user, real time (seconds) */
double atm_cells, atm_speed;

void err();
void mes();
int pattern();
void prep_timer();
double read_timer();
int Nread();
int Nwrite();
void delay();
int mread();
char *outfmt();
int getopt(int argc, char *argv[], char *optstring);

void
sigpipe()
{
}

main(argc,argv)
int argc;
char **argv;
{
        unsigned long addr_tmp;
        int c;
        WORD wVersionRequested; 
        WSADATA wsaData; 
        int error; 

        if (argc < 2) goto usage;

        while ((c = getopt(argc, argv, "drstuvBDTb:f:l:n:p:A:O:")) != -1) {
                switch (c) {

                case 'B':
                        b_flag = 1;
                        break;
                case 't':
                        trans = 1;
                        break;
                case 'r':
                        trans = 0;
                        break;
                case 'd':
                        options |= SO_DEBUG;
                        break;
                case 'D':
#ifdef TCP_NODELAY
                        nodelay = 1;
#else
                        fprintf(stderr, 
        "ttcp: -D option ignored: TCP_NODELAY socket option not supported\n");
#endif
                        break;
                case 'n':
                        nbuf = atoi(optarg);
                        break;
                case 'l':
                        buflen = atoi(optarg);
                        break;
                case 's':
                        sinkmode = !sinkmode;
                        break;
                case 'p':
                        port = atoi(optarg);
                        break;
                case 'u':
                        udp = 1;
                        break;
                case 'v':
#ifdef WIN32
                        fprintf(stderr, 
                            "ttcp: -v option ignored: Verbose mode calls not supported by WIN32.\n");
#else
                        verbose = 1;
                        break;
#endif //WIN32
                case 'M':
                        atm = 1;
                        break;
                case 'A':
                        bufalign = atoi(optarg);
                        break;
                case 'O':
                        bufoffset = atoi(optarg);
                        break;
                case 'b':
#if defined(SO_SNDBUF) || defined(SO_RCVBUF)
                        sockbufsize = atoi(optarg);
#else
                        fprintf(stderr, 
"ttcp: -b option ignored: SO_SNDBUF/SO_RCVBUF socket options not supported\n");
#endif
                        break;
                case 'f':
                        fmt = *optarg;
                        break;
                case 'T':
                        touchdata = 1;
                        break;

                default:
                        goto usage;
                }
        }
  
        wVersionRequested = MAKEWORD(1, 1); 
 
        /* Initialize the WINSOCK.DLL 
        */
        error = WSAStartup(wVersionRequested, &wsaData); 
 
        if (error != 0) 
            /* Tell the user that we couldn't find a useable */ 
            /* winsock.dll.     */ 
            return 0; 
 
        /* Confirm that the Windows Sockets DLL supports 1.1.*/ 
        /* Note that if the DLL supports versions greater */ 
        /* than 1.1 in addition to 1.1, it will still return */ 
        /* 1.1 in wVersion since that is the version we */ 
        /* requested. */ 
 
        if ( LOBYTE( wsaData.wVersion ) != 1 || 
                HIBYTE( wsaData.wVersion ) != 1 ) { 
            /* Tell the user that we couldn't find a useable */ 
            /* winsock.dll. */ 
            WSACleanup(); 
            return; 
        } 
        /* The Windows Sockets DLL is acceptable. Proceed. */ 
 
        sinme.sin_family = AF_INET;
        if(trans)  {
                /* xmitr */
                if (optind == argc)
                        goto usage;
                memset((char *)&sinhim, 0, sizeof(sinhim));
                host = argv[optind];
                if (atoi(host) > 0 )  {
                        /* Numeric */
                        sinhim.sin_family = AF_INET;
#if defined(cray)
                        addr_tmp = inet_addr(host);
                        sinhim.sin_addr = addr_tmp;
#else
                        sinhim.sin_addr.s_addr = inet_addr(host);
#endif
                } else {
                        if ((addr=gethostbyname(host)) == NULL)
                                err("bad hostname");
                        sinhim.sin_family = addr->h_addrtype;
                        CopyMemory(&addr_tmp,(const char*)addr->h_addr, sizeof(addr_tmp));                    
#if defined(cray)
                        sinhim.sin_addr = addr_tmp;
#else
                        sinhim.sin_addr.s_addr = addr_tmp;
#endif /* cray */
                }
                sinhim.sin_port = htons(port);
                sinme.sin_port = 0;             /* free choice */
        } else {
                /* rcvr */
                sinme.sin_port =  htons(port);
        }


        if (udp && buflen < 5) {
            buflen = 5;         /* send more than the sentinel size */
        }

        if ( (buf = (char *)malloc(buflen+bufalign)) == (char *)NULL)
                err("malloc");
        if (bufalign != 0)
                buf +=(bufalign - ((int)buf % bufalign) + bufoffset) % bufalign;

        if (trans) {
            fprintf(stdout,
            "ttcp-t: buflen=%d, nbuf=%d, align=%d/%d, port=%d",
                buflen, nbuf, bufalign, bufoffset, port);
            if (sockbufsize)
                fprintf(stdout, ", sockbufsize=%d", sockbufsize);
            fprintf(stdout, "  %s  -> %s\n", udp?"udp":"tcp", host);
        } else {
            fprintf(stdout,
            "ttcp-r: buflen=%d, nbuf=%d, align=%d/%d, port=%d",
                buflen, nbuf, bufalign, bufoffset, port);
            if (sockbufsize)
                fprintf(stdout, ", sockbufsize=%d", sockbufsize);
            fprintf(stdout, "  %s\n", udp?"udp":"tcp");
        }

        if ((fd = socket(AF_INET, udp?SOCK_DGRAM:SOCK_STREAM, 0)) < 0)
                err("socket");
        mes("socket");

        if (bind(fd, (struct sockaddr *) &sinme, sizeof(sinme)) < 0)
                err("bind");

#if defined(SO_SNDBUF) || defined(SO_RCVBUF)
        if (sockbufsize) {
            if (trans) {
                if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sockbufsize,
                    sizeof sockbufsize) < 0)
                        err("setsockopt: sndbuf");
                mes("sndbuf");
            } else {
                if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &sockbufsize,
                    sizeof sockbufsize) < 0)
                        err("setsockopt: rcvbuf");
                mes("rcvbuf");
            }
        }
#endif
        if (!udp)  {
#if (0)
            signal(SIGPIPE, sigpipe);
#endif
            if (trans) {
                /* We are the client if transmitting */
                if (options)  {
#if defined(BSD42)
                        if( setsockopt(fd, SOL_SOCKET, options, 0, 0) < 0)
#else /* BSD43 */
                        if( setsockopt(fd, SOL_SOCKET, options, &one, sizeof(one)) < 0)
#endif
                                err("setsockopt");
                }
#ifdef TCP_NODELAY
                if (nodelay) {
                        struct protoent *p;
                        p = getprotobyname("tcp");
                        if( p && setsockopt(fd, p->p_proto, TCP_NODELAY, 
                            &one, sizeof(one)) < 0)
                                err("setsockopt: nodelay");
                        mes("nodelay");
                }
#endif
                if(connect(fd, (struct sockaddr *)&sinhim, sizeof(sinhim)) < 0)
                        err("connect");
                mes("connect");
            } else {
                /* otherwise, we are the server and 
                 * should listen for the connections
                 */
                listen(fd,0);   /* allow a queue of 0 */
                if(options)  {
#if defined(BSD42)
                        if( setsockopt(fd, SOL_SOCKET, options, 0, 0) < 0)
#else /* BSD43 */
                        if( setsockopt(fd, SOL_SOCKET, options, &one, sizeof(one)) < 0)
#endif
                                err("setsockopt");
                }
                fromlen = sizeof(frominet);
                domain = AF_INET;
                if((fd=accept(fd, (struct sockaddr *)&frominet, &fromlen)) < 0)
                        err("accept");
                { struct sockaddr_in peer;
                  int peerlen = sizeof(peer);
                  if (getpeername(fd, (struct sockaddr *) &peer, 
                                &peerlen) < 0) {
                        err("getpeername");
                  }
                  fprintf(stderr,"ttcp-r: accept from %s\n", 
                        inet_ntoa(peer.sin_addr));
                }
            }
        }
        prep_timer();
        errno = 0;
        if (sinkmode) {      
                register int cnt;
                if (trans)  {
                        pattern( buf, buflen );
                        if(udp)  (void)Nwrite( fd, buf, 4 ); /* rcvr start */
                        while (nbuf-- && Nwrite(fd,buf,buflen) == buflen)
                                nbytes += buflen;
                        if(udp)  (void)Nwrite( fd, buf, 4 ); /* rcvr end */
                } else {
                        if (udp) {
                            while ((cnt=Nread(fd,buf,buflen)) > 0)  {
                                    static int going = 0;
                                    if( cnt <= 4 )  {
                                            if( going )
                                                    break;      /* "EOF" */
                                            going = 1;
                                            prep_timer();
                                    } else {
                                            nbytes += cnt;
                                    }
                            }
                        } else {
                            while ((cnt=Nread(fd,buf,buflen)) > 0)  {
                                    nbytes += cnt;
                            }
                        }
                }
        } else {
                register int cnt;
                if (trans)  {
                        while((cnt=_read(0,buf,buflen)) > 0 &&
                            Nwrite(fd,buf,cnt) == cnt)
                                nbytes += cnt;
                }  else  {
                        while((cnt=Nread(fd,buf,buflen,0)) > 0 &&
                            _write(1,buf,cnt) == cnt)
                                nbytes += cnt;
                }
        }
        if(errno) err("IO");
        (void)read_timer(stats,sizeof(stats));
        if(udp&&trans)  {
                (void)Nwrite( fd, buf, 4 ); /* rcvr end */
                (void)Nwrite( fd, buf, 4 ); /* rcvr end */
                (void)Nwrite( fd, buf, 4 ); /* rcvr end */
                (void)Nwrite( fd, buf, 4 ); /* rcvr end */
        }
        if( elapsed_time <= 0.0 )  elapsed_time = 0.001;
        if(udp)
        {
          atm_cells = (((double)nbytes+(double)nbuf*28)/48.0);
          atm_speed = atm_cells*53.0*8/elapsed_time;
        }
        else
        {
          atm_cells = (((double)nbytes+(double)nbuf*40)/48.0);
          atm_speed = atm_cells*53.0*8/elapsed_time;
          
        }

        fprintf(stdout,
                "ttcp%s: %ld bytes in %.2f seconds = %s/sec +++\n",
                trans?"-t":"-r",
                nbytes, elapsed_time, outfmt(((double)nbytes)/elapsed_time));
        if (atm)
        {
            fprintf(stdout,
                "ttcp%s: Which is at least %.1f ATM cells or %0.2f Mbits/sec at adapter level\n",
                trans?"-t":"-r", atm_cells, atm_speed/1000000);
        }
        if (verbose) {
            fprintf(stdout,
                "ttcp%s: %ld bytes in %.2f CPU seconds = %s/sec\n",
                trans?"-t":"-r",
                nbytes, elapsed_time, outfmt(((double)nbytes)/elapsed_time));
        }
        fprintf(stdout,
                "ttcp%s: %d I/O calls, msec/call = %.2f, calls/sec = %.2f\n",
                trans?"-t":"-r",
                numCalls,
                1024.0 * elapsed_time/((double)numCalls),
                ((double)numCalls)/elapsed_time);
        fprintf(stdout,"ttcp%s: %s\n", trans?"-t":"-r", stats);
        if (verbose) {
            fprintf(stdout,
                "ttcp%s: buffer address %#x\n",
                trans?"-t":"-r",
                buf);
        }
        closesocket(fd);
        WSACleanup(); 
        exit(0);

usage:
        fprintf(stderr,Usage);
        
        /* Deinitialize the WINSOCK.DLL 
        */
        WSACleanup(); 
        exit(1);
}

void
err(s)
char *s;
{
    long error;
        fprintf(stderr,"ttcp%s: ", trans?"-t":"-r");
        error = WSAGetLastError();
        fprintf(stderr,"error=%d\n",error);
        WSACleanup(); 
        exit(1);
}

void
mes(s)
char *s;
{
        fprintf(stderr,"ttcp%s: %s\n", trans?"-t":"-r", s);
}

pattern( cp, cnt )
register char *cp;
register int cnt;
{
        register char c;
        c = 0;
        while( cnt-- > 0 )  {
                while( !isprint((c&0x7F)) )  c++;
                *cp++ = (c++&0x7F);
        }
}

char *
outfmt(b)
double b;
{
    static char obuf[50];
    switch (fmt) {
        case 'G':
            sprintf(obuf, "%.2f GB", b / 1024.0 / 1024.0 / 1024.0);
            break;
        default:
        case 'K':
            sprintf(obuf, "%.2f KB", b / 1024.0);
            break;
        case 'M':
            sprintf(obuf, "%.2f MB", b / 1024.0 / 1024.0);
            break;
        case 'g':
            sprintf(obuf, "%.2f Gbit", b * 8.0 / 1024.0 / 1024.0 / 1024.0);
            break;
        case 'k':
            sprintf(obuf, "%.2f Kbit", b * 8.0 / 1024.0);
            break;
        case 'm':
            sprintf(obuf, "%.2f Mbit", b * 8.0 / 1024.0 / 1024.0);
            break;
        case 'x':
            sprintf(obuf, "%.2f kilobit", b * 8.0 / 1000.0);
            break;
        case 'y':
            sprintf(obuf, "%.2f megabit", b * 8.0 / 1000.0 / 1000.0);
            break;
        case 'z':
            sprintf(obuf, "%.2f gigabit", b * 8.0 / 1000.0 / 1000.0 / 1000.0);
            break;
    }
    return obuf;
}

time_t time0;  /* Time at which timing started */

/*
 *                      P R E P _ T I M E R
 */
void
prep_timer()
{
    time0 = clock();
}

/*
 *                      R E A D _ T I M E R
 * 
 */
double
read_timer(str,len)
char *str;
{
        elapsed_time = (double)(clock() - time0) / 1000;
        return elapsed_time;
}


/*
 *                      N R E A D
 */
Nread( fd, buf, count )
int fd;
void *buf;
int count;
{
        struct sockaddr_in from;
        int len = sizeof(from);
        register int cnt;
        if( udp )  {
                cnt = recvfrom( fd, buf, count, 0,(struct sockaddr *)&from,
                               &len );
                numCalls++;
        } else {
                if( b_flag )
                        cnt = mread( fd, buf, count );  /* fill buf */
                else {
                        cnt = recv( fd, buf, count, 0 );
                        numCalls++;
                }
                if (touchdata && cnt > 0) {
                        register int c = cnt, sum;
                        register char *b = buf;
                        while (c--)
                                sum += *b++;
                }
        }
        return(cnt);
}

/*
 *                      N W R I T E
 */
Nwrite( fd, buf, count )
int fd;
void *buf;
int count;
{
        register int cnt;
        if( udp )  {
again:
                cnt = sendto( fd, buf, count, 0, (struct sockaddr *) &sinhim,
                             sizeof(sinhim) );
                numCalls++;
                if( cnt == SOCKET_ERROR && errno == WSAENOBUFS )  {
                        delay(18000);
                        errno = 0;
                        goto again;
                }
        } else {
                cnt = send( fd, buf, count, 0 );
                numCalls++;
        }
        return(cnt);
}

void
delay(us)
{
        struct timeval tv;

        tv.tv_sec = 0;
        tv.tv_usec = us;
        (void)select( 1, (char *)0, (char *)0, (char *)0, &tv );
}

/*
 *                      M R E A D
 *
 * This function performs the function of a read(II) but will
 * call read(II) multiple times in order to get the requested
 * number of characters.  This can be necessary because
 * network connections don't deliver data with the same
 * grouping as it is written with.  Written by Robert S. Miles, BRL.
 */
int
mread(fd, bufp, n)
int fd;
register char   *bufp;
unsigned        n;
{
        register unsigned       count = 0;
        register int            nread;

        do {
                nread = recv(fd, bufp, n-count, 0);
                numCalls++;
                if(nread < 0)  {
                        perror("ttcp_mread");
                        return(-1);
                }
                if(nread == 0)
                        return((int)count);
                count += (unsigned)nread;
                bufp += nread;
         } while(count < n);

        return((int)count);
}

int getopt(int argc, char *argv[], char *optstring)
{
    int option;
    char *p;

    if (optind == argc)
        return -1;
    
    p = argv[optind];
    if (*p != '-')
        return -1;

    option = *(++p);
    p++;
    if (*p != '\0')
        optarg = p;
    else
        optarg = argv[optind + 1];
    
    /* increment count for next function call */                   

    optind++;        
    return option;
}